
# The key of the idea here: create an object that is an in-place gradient that
# checks for compiled gradient tapes, and if it doesn't have one it generates
# it and then uses that. Then the next time you call, it has that tape handy.
struct GradWithTape{F} <: Function
  f::F
  tapes::Dict{Type, ReverseDiff.CompiledTape}
end


struct FoRWrapper{T,F,C}
  arglen::Int64 
  gradhash::Ref{UInt64}
  hesshash::Ref{UInt64}
  buf_g::Vector{T}
  buf_h::Matrix{T}
  fun::F
  grad_withtape::GradWithTape{F}
  cfg::C
end


function GradWithTape(f)
  GradWithTape(f, Dict{Type,ReverseDiff.CompiledTape}())
end


function (gt::GradWithTape{F})(g, p::P) where{F,P}
  if !haskey(gt.tapes, P)
    newtape     = ReverseDiff.GradientTape(gt.f, p)
    cnewtape    = ReverseDiff.compile(newtape)
    gt.tapes[P] = cnewtape
  end
  ReverseDiff.gradient!(g, gt.tapes[P], p)
end


function FoRWrapper(f::F, sample_arg::Vector{T}, 
                    chunksize=min(length(sample_arg),16)) where{F,T}
  # Allocate everything:
  n = length(sample_arg)
  buf_g    = Array{T}(undef, n)
  buf_h    = Array{T}(undef, n, n)
  gradhash = Ref(hash(buf_g))
  hesshash = Ref(hash(buf_h))
  grad_withtape = GradWithTape(f)
  # A jacobian config to cut down on allocs in the final calling interface.
  cfg = ForwardDiff.JacobianConfig(grad_withtape, buf_g, sample_arg, 
                                   ForwardDiff.Chunk{chunksize}())
  FoRWrapper(n, gradhash, hesshash, buf_g, buf_h, f, grad_withtape, cfg)
end


function (M::FoRWrapper{T,F,C})(case::Symbol, arg::Vector{T}, 
                                checkhash=true) where{T,F,C}
  # Check lengths:
  @assert length(arg) == M.arglen "Provided argument is not the correct length."
  # Check if we happen to have the desired grad or hessian sitting around:
  arghash = hash(arg)
  if checkhash
    ((case == :gradient) && (arghash === M.gradhash[])) && return M.buf_g
    ((case == :hessian)  && (arghash === M.hesshash[])) && return M.buf_h
  end
  # Now give back either value, gradient, or hessian, updating the cache value
  # appropriately for either one:
  if case == :value
    return M.fun(arg)
  elseif case == :gradient
    M.grad_withtape(M.buf_g, arg)
    M.gradhash[] = arghash
    return M.buf_g
  elseif case == :hessian
    ForwardDiff.jacobian!(M.buf_h, M.grad_withtape, M.buf_g, arg, M.cfg)
    M.gradhash[] = arghash
    M.hesshash[] = arghash
    return M.buf_h
  else
    throw(error("Case options are :value, :gradient, or :hessian."))
  end
end

